-- -------------------------------------------------------------------------------------
--
-- DruidFunc
--     Version 2.12b
--
-- -------------------------------------------------------------------------------------
--
-- Original Source by Gabraelle
-- http://ruined-by-experience.com/nocturnum/DruidFunc.html  (Defunct link)
--
-- Incorporated some functionality by Adora (Detect max spell rank)
-- http://www.gerdofal.net/wow/druid_mod.html
--
-- This update by Patrick Hunlock  AkA Amid (Hyjal-alliance/Dark Iron-horde) (Version 2.0 forward)
-- http://www.hunlock.com/df.php
--
-- -------------------------------------------------------------------------------------
--
-- Version History:
--   6/20/06- v2.12b- Fixed targeting errors in rejuvenation (in regards to swiftmend)
--                  - Fixed a text error when saving your bear outfit.
--                  - Rejuvenation does low mana checks for weapon swaps now
--                  - Function Rejuv added which restores original rejuvenation functionality
--                  -    (for those that don't like the enhanced rejuvenation)
--                  - Rejuvenation warnings will now play a short sound to indicate rejuvenation didn't actually cast.
--                  - Fixed errors when asking rejuvenation and regrowth to be chatty.
--   6/04/06- v2.12a- Fix for dismount code when requesting shapeshift (code by Bitbyte of Icecrown -- THANKS!)
--                  - Innervate aborts if target already has innervate (code by Bitbyte of Icecrown -- THANKS!)
--                  - ManaMe() now calls Druid.Innervate (after setting player as the target)
--                  -   This provides proper target checking (too much mana, already has innervate, etc)
--                  - Fix to +healing and mana/5 checks by Shreavel (Curse gaming -- THANKS!)
--                  - If user is in a raid, healing spells will no longer auto select the user if the target is blank or an enemy
--                  -    By allowing the spell to stay in targeting mode the "healrange" mod will work properly if installed
--                  -    Healrange mod: http://ui.worldofwar.net/ui.php?id=1769
--                  - Rejuvenation rules have been altered.
--                  -   If you have swiftmend, casting rejuvenation on a player that already has rejuvenation will cast swiftmend instead.
--                  -   If swiftmend is on cooldown or you don't have swiftmend then
--                  -     you will get a warning that the player already has rejuvenation and the spell will not cast.
--                  -     You will have to tap the button a second time to force a recast.
--                  -   Fixed a bug where the rejuvenation check wouldn't work on you, just your targets ( Thanks Kagar of Azgalor )
--                  - /DPlusheal will now automatically add in the Zandalar Hero Charm and the Hibernation Crystal when they are active.
--                  - If you have auto trinkets enabled hibernation crystal will attempt to be used on every heal spell.
--                  - Added new slash command /DTrinkets [on/off]
--                  -    When on the following trinkets will attempt to automatically be used...
--                  -       Mar'li's Eye whenever you've used 600 mana
--                  -       Hibernation Crystal whenever you attempt to cast a healing spell
--                  -       Zandalar Hero Charm whenever you attempt to cast an offensive spell.
--                  -       Rune of metamorphasis whenever you attempt to shapeshift.
--                  -       Defiler's Talisman whenever you hit barkskin
--                  - Hibernate will no longer announce that you are casting if the target is already slept.
--                  - You can now pass a minumum rank to PowerTouch (Before it always defaulted to rank 1)  ( Thanks Kagar of Azgalor )
--   5/04/06- v2.12 - /DPlusHeal will now also show you mana per 5 from equipped gear (for the curious)
--                  - /DPlusHeal will now show you your spirit regen (outside 5 sec rule -- for the curious)
--                  - /DPlusHeal will now count set bonuses only once instead of for each item.
--                  - /DWeapon has been removed and replaced with /DOutfit
--                  -   /DOutfit does complete wardrobe changes on shift from head to trinkets if enabled.
--                  -   You must enable complete wardrobe changes with /Doutfit outfit, otherwise only weapons change.
--                  -   Equip what you want to wear then type /doutfit (form) to save that outfit for that form
--                  -   /DOutfit weapon (will swap only weapons/idols -- 1.11 functionality)  
--                  -   /DOutfit outfit (will swap all gear from head to toe)
--                  -   /DOutfit off  (disable wardrobe changes)
--                  -   /DOutfit on   (enable wardrobe changes)
--                  -   /DOutfit [cat/bear/moon/travel/aqua/human/mana/innervate/mount] for respective forms
--                  -      human is base wardrobe to equip when shifting out of form and not in combat
--                  -      mana equipes the desired weapons when your mana falls below 85%
--                  -      innervate equips the desired weapons when you innervate.
--                  -      mount is your riding gear
--                  - Added /DManaC [silent/announce]  
--                  -    Silent will prevent spell abort messages from being broadcast to you
--                  -    Announce will re-enable announcements in the default chat window
--                  -    Default is announce
--                  - Added keybind and macro ManaMe() this will target the player and cast innervate (self innervate only)
--                  - Innervate has been reworked.
--                  -    Can be cast on any target with a mana pool
--                  -    If target is enemy or target has no mana pool the spell will target you.
--                  -    If the target has more than 50% mana the spell will abort
--                  - Fixed typo in find maxrank for demoralizing roar.
--                  - Desperate Prayer will only attempt a NS heal if you are below 50% health.
--   4/19/06- v2.11 - If hardshifting is on when the shift button is pressed at the prowl stage
--                  -   detect humanoid will automatically be activated.
--                  - If hardshifting is on when the shift button is pressed for the ravage stage,
--                  -   tiger's fury will automatically be activated IF you have the energy for both
--                  -   tiger's fury and ravage, if you do not then only ravage will be attempted.
--                  - /dplusheal auto will now automatically calculate your +heal based on the items you are wearing.
--                  -   You can still use /dplusheal # to manually force an offset and you can use a negative number
--                  -   to force a higher rank of healing touch to be used.
--                  - Shapeshift requests will be ignored if you're not a druid (just in case you have a warrior/paladin
--                  -   and forgot to disable this mod for those characters like a certain mod author).
--                  - Automatic Weapon swaps on shifting now happen on the shift, instead of on a request.
--                  -   This prevents weapons from being swapped when you actually weren't able to enter that form.
--                  - Added "power touch" keybind and macro (Druid.PowerTouch()) which will attempt to 
--                  -   activate a zandalarian hero charm and wooloshay's charm of nature to give you the
--                  -   fastest and most powerful heals available if you're wearing your trinkets.   The
--                  -   trinkets will be activated and then healing touch will be called.
--                  -   I use "2" for healing touch, and "alt-2" for power touch.  It's a nice panic system.
--                  - Healing touch now accepts an optional forth paramater (forceRank), if you pass true
--                  -   here then healing touch will cast ONLY the spell rank you requested in minRank.  This is
--                  -   is for people who want to only cast healing touch rank 4 but still utilize the mana conserve
--                  -   features of this mod.  So to only cast rank 4 you'd set up a macro as such...
--                  -      /script Druid.HealingTouch(4, false, false, true)
--                  - Desperate prayer will now ALWAYS attempt to use nature's swiftness (regardless of your health)
--                  -   This to keep the panic spell consistant so you always know what to expect.
--                  - If you are in a feral form Desperate prayer will shift you back to humanoid, you will, however,
--                  -    need to tap desperate prayer again to get the heal.
--                  - When shifting into moonkin form, the actionbar will automatically be set to the 5th page
--                  -   This gives you a way to set up a moonkin specific action bar the same as if you'd set up
--                  -   bear or cat.
--                  - Fixed a bug preventing faerie fire from casting in feral form (if you have the skill)
--                  - Fixed Feral Charge keybind and macro.
--                  - If hardshifted & bear, pressing the bear shift key again will attempt a feral charge
--                  -    otherwise will maul.
--                  - TestTrinket is no longer case depenedant (ZaNDalar HERO charm will still find the charm)
--                  - Added Macro/Keybind BestAttackCat
--                  -   If not in combat will attempt to Track Humanoids & Stealth
--                  -   If Stealthed will attempt tiger's fury & Ravage (if enough energy, otherwise only Ravage)
--                  -   If not stealthed will claw (Will attempt Faerie Fire (feral) if not enough energy).
--                  -   If 5 combo points will ferocious bite.
--                  - Fixed a bug preventing human offhand weapons from equipping (again)
-- -------------------------------------------------------------------------------------

-- -------------------------------------------------------------------------------------
-- These variables extend WoW's keyboard definition panel (Esc, keybindings ingame)
-- They work hand in hand with the bindings.xlm file in this directory.
-- -------------------------------------------------------------------------------------

   BINDING_HEADER_DFUNC                           = "DruidFunc Library";
   BINDING_NAME_DFUNC_DesperatePrayer             = "Desperate Prayer";
   BINDING_NAME_DFUNC_FaerieFire                  = 'Faerie Fire';
   BINDING_NAME_DFUNC_Hibernate                   = 'Hibernate';
   BINDING_NAME_DFUNC_Moonfire                    = 'Moonfire';
   BINDING_NAME_DFUNC_Starfire                    = 'Starfire';
   BINDING_NAME_DFUNC_Hurricane                   = "Hurricane";
   BINDING_NAME_DFUNC_Wrath                       = 'Wrath';
   BINDING_NAME_DFUNC_MotW                        = 'Mark of the Wild';
   BINDING_NAME_DFUNC_Thorns                      = 'Thorns';
   BINDING_NAME_DFUNC_Roots                       = "Entangling Roots";
   BINDING_NAME_DFUNC_NaturesGrasp                = "Nature's Grasp";
   BINDING_NAME_DFUNC_AbolishPoison               = 'Abolish Poison';
   BINDING_NAME_DFUNC_Tranquility                 = "Tranquility";
   BINDING_NAME_DFUNC_Regrowth                    = 'Regrowth';
   BINDING_NAME_DFUNC_Rejuvenation                = 'Rejuvenation';
   BINDING_NAME_DFUNC_RemoveCurse                 = 'Remove Curse';
   BINDING_NAME_DFUNC_HealingTouch                = 'Healing Touch';
   BINDING_NAME_DFUNC_Barkskin                    = 'Barkskin';
   BINDING_NAME_DFUNC_Innervate                   = 'Innervate';
   BINDING_NAME_DFUNC_ManaMe                      = "Mana Me";
   BINDING_NAME_DFUNC_Human                       = 'Humanoid Form';
   BINDING_NAME_DFUNC_Bear                        = 'Bear Form';
   BINDING_NAME_DFUNC_Cat                         = 'Cat Form';
   BINDING_NAME_DFUNC_Travel                      = 'Travel Form';
   BINDING_NAME_DFUNC_Aqua                        = 'Aquatic Form';
   BINDING_NAME_DFUNC_Moon                        = 'Moonkin Form';
   BINDING_NAME_DFUNC_InsectSwarm                 = 'Insect Swarm';
   BINDING_NAME_DFUNC_Rebirth                     = 'Rebirth';
   BINDING_NAME_DFUNC_Ravage                      = "Ravage";
   BINDING_NAME_DFUNC_Shred                       = "Shred";
   BINDING_NAME_DFUNC_Claw                        = "Claw";
   BINDING_NAME_DFUNC_Rake                        = "Rake";
   BINDING_NAME_DFUNC_Rip                         = "Rip";
   BINDING_NAME_DFUNC_Pounce                      = "Pounce";
   BINDING_NAME_DFUNC_Prowl                       = "Prowl";
   BINDING_NAME_DFUNC_Cower                       = "Cower";
   BINDING_NAME_DFUNC_Dash                        = "Dash";
   BINDING_NAME_DFUNC_FerociousBite               = "Ferocious Bite";
   BINDING_NAME_DFUNC_TigersFury                  = "Tiger's Fury";
   BINDING_NAME_DFUNC_FeralCharge                 = "Feral Charge";
   BINDING_NAME_DFUNC_Maul                        = "Maul";
   BINDING_NAME_DFUNC_Bash                        = "Bash";
   BINDING_NAME_DFUNC_Growl                       = "Growl";
   BINDING_NAME_DFUNC_Swipe                       = "Swipe";
   BINDING_NAME_DFUNC_FrenziedRegen               = "Frenzied Regeneration";
   BINDING_NAME_DFUNC_DemRoar                     = "Demoralizing Roar";
   BINDING_NAME_DFUNC_ChalRoar                    = "Challenging Roar";
   BINDING_NAME_DFUNC_Enrage                      = "Enrage";
   BINDING_NAME_DFUNC_PowerTouch                  = "Power Touch";
   BINDING_NAME_DFUNC_BestCat                     = "Best Attack Cat";

-----------------------------------------------------------------------------------   
--These variables are registered for save by character in the .TOC file
-----------------------------------------------------------------------------------   

DFUNCPlusHeal   = 0;            -- The value of +heal from items
DFUNCHardShift  = "off";        -- Hardshifting functionality
DFUNCManaC      = "off";        -- Either "off" or a number which is the abort percentage
DFUNCSwap       = true;         -- True if weapon/outfit swaps are enabled.
DFUNCSwiftTarget= true;         -- Changes the function of swiftshift keypresses
DFUNCSilent     = false;        -- True keeps mana conserve messages out of chat.
DFUNCWeapon     = false;        -- True if only swap weapons/false to swap everything
DFUNCTrinkets   = false;


DFUNCHumanSet = {}              -- An array which holds all of the inventory items for the humanoid set
DFUNCCatSet   = {}              -- An array which holds all of the inventory items for the cat form set
DFUNCBearSet  = {}              -- An array which holds all of the inventory items for the bear form set
DFUNCMoonSet  = {}              -- An array which holds all of the inventory items for the moonkin form set (heh)
DFUNCTravelSet= {}              -- An array which holds all of the inventory items for the travel form set.
DFUNCAquaSet  = {}              -- An array which holds all of the inventory items for the aquatic form set
DFUNCManaSet  = {}              -- An array which holds all of the inventory items for the low mana set.
DFUNCISet     = {}              -- An array which holds all of the inventory items for the innervate set.
DFUNCMountSet = {}              -- An array which holds all of the inventory items for the player's mount set.

-----------------------------------------------------------------------------------   
-- Global variables (not saved)
-----------------------------------------------------------------------------------   

DFUNCVersion   = 1.12              -- Version number and identifier so other mods can test for this function

DFUNC_Slots    = {"Head","Neck","Shoulder","Shirt","Chest","Waist","Legs","Feet","Wrist","Hands","Finger0","Finger1","Trinket0","Trinket1","Back","MainHand","SecondaryHand","Ranged","Tabard"};

DFUNCLastForm   = "humanoid form"   -- Last form the player was in.
DFUNCNoScan     = false             -- True if inventory checks should be disabled
DFUNCIsMounted  = false             -- True if player is on a mount
DFUNCInnervate  = false             -- True during Innervates
DFUNCRejuvWarn  = ""                -- Last player that got a "rejuvenation already active" warning
DFUNCTrinketHeal= 0                 -- Heal modifier from activated trinkets

-----------------------------------------------------------------------------------   
--Begin Druid Func
-----------------------------------------------------------------------------------   

  
Druid = {

-- -------------------------------------------------------------------------------------
-- These functions setup initial variables and states.
-- -------------------------------------------------------------------------------------

   SetAllMaxRanks = function()
      -- -------------------------------------------------------------------------------------
      -- Find the maximum ranks for the various spells.
      -- ---------------------------------------------- ---------------------------------------
      MAXRANKNAME_HealingTouch,  MAXRANK_HealingTouch,  BOOKINDEX_HealingTouch   = Druid.maxRank("Healing Touch");
      MAXRANKNAME_Rejuv,         MAXRANK_Rejuv,         BOOKINDEX_Rejuv          = Druid.maxRank("Rejuvenation");
      MAXRANKNAME_Regrow,        MAXRANK_Regrow,        BOOKINDEX_Regrow         = Druid.maxRank("Regrowth");
      MAXRANKNAME_Thorns,        MAXRANK_Thorns,        BOOKINDEX_Thorns         = Druid.maxRank("Thorns");
      MAXRANKNAME_MoTW,          MAXRANK_MoTW,          BOOKINDEX_MoTW           = Druid.maxRank("Mark of the Wild");
      MAXRANKNAME_Moonfire,      MAXRANK_Moonfire,      BOOKINDEX_Moonfire       = Druid.maxRank("Moonfire");
      MAXRANKNAME_Starfire,      MAXRANK_Starfire,      BOOKINDEX_Starfire       = Druid.maxRank("Starfire");
      MAXRANKNAME_Wrath,         MAXRANK_Wrath,         BOOKINDEX_Wrath          = Druid.maxRank("Wrath");
      MAXRANKNAME_InsectSwarm,   MAXRANK_InsectSwarm,   BOOKINDEX_InsectSwarm    = Druid.maxRank("Insect Swarm");
      MAXRANKNAME_FF,            MAXRANK_FF,            BOOKINDEX_FF             = Druid.maxRank("Faerie Fire");
      MAXRANKNAME_FFF,           MAXRANK_FFF,           BOOKINDEX_FFF            = Druid.maxRank("Faerie Fire (Feral)");
      MAXRANKNAME_Hibernate,     MAXRANK_Hibernate,     BOOKINDEX_Hibernate      = Druid.maxRank("Hibernate");
      MAXRANKNAME_Rebirth,       MAXRANK_Rebirth,       BOOKINDEX_Rebirth        = Druid.maxRank("Rebirth");
      MAXRANKNAME_NaturesGrasp,  MAXRANK_NaturesGrasp,  BOOKINDEX_NaturesGrasp   = Druid.maxRank("Nature's Grasp");
      MAXRANKNAME_Hurricane,     MAXRANK_Hurricane,     BOOKINDEX_Hurricane      = Druid.maxRank("Hurricane");
      MAXRANKNAME_Tranquility,   MAXRANK_Tranquility,   BOOKINDEX_Tranquility    = Druid.maxRank("Tranquility");
      MAXRANKNAME_Roots,         MAXRANK_Roots,         BOOKINDEX_Roots          = Druid.maxRank("Entangling Roots");
      MAXRANKNAME_Ravage,        MAXRANK_Ravage,        BOOKINDEX_Ravage         = Druid.maxRank("Ravage");
      MAXRANKNAME_Shred,         MAXRANK_Shred,         BOOKINDEX_Shred          = Druid.maxRank("Shred");
      MAXRANKNAME_Claw,          MAXRANK_Claw,          BOOKINDEX_Claw           = Druid.maxRank("Claw");
      MAXRANKNAME_Prowl,         MAXRANK_Prowl,         BOOKINDEX_Prowl          = Druid.maxRank("Prowl");
      MAXRANKNAME_Rake,          MAXRANK_Rake,          BOOKINDEX_Rake           = Druid.maxRank("Rake");
      MAXRANKNAME_Pounce,        MAXRANK_Pounce,        BOOKINDEX_Pounce         = Druid.maxRank("Pounce");
      MAXRANKNAME_Cower ,        MAXRANK_Cower,         BOOKINDEX_Cower          = Druid.maxRank("Cower");
      MAXRANKNAME_Dash,          MAXRANK_Dash,          BOOKINDEX_Dash           = Druid.maxRank("Dash");
      MAXRANKNAME_Rip,           MAXRANK_Rip,           BOOKINDEX_Rip            = Druid.maxRank("Rip");
      MAXRANKNAME_FerociousBite, MAXRANK_FerociousBite, BOOKINDEX_FerociousBite  = Druid.maxRank("Ferocious Bite");
      MAXRANKNAME_FeralCharge,   MAXRANK_FeralCharge,   BOOKINDEX_FeralCharge    = Druid.maxRank("Feral Charge");
      MAXRANKNAME_TigersFury,    MAXRANK_TigersFury,    BOOKINDEX_TigersFury     = Druid.maxRank("Tiger's Fury");
      MAXRANKNAME_Maul,          MAXRANK_Maul,          BOOKINDEX_Maul           = Druid.maxRank("Maul");
      MAXRANKNAME_Bash,          MAXRANK_Bash,          BOOKINDEX_Bash           = Druid.maxRank("Bash");
      MAXRANKNAME_Swipe,         MAXRANK_Swipe,         BOOKINDEX_Swipe          = Druid.maxRank("Swipe");
      MAXRANKNAME_FrenziedRegen, MAXRANK_FrenziedRegen, BOOKINDEX_FrenziedRegen  = Druid.maxRank("Frenzied Regeneration");
      MAXRANKNAME_DemRoar,       MAXRANK_DemRoar,       BOOKINDEX_DemRoar        = Druid.maxRank("Demoralizing Roar");
      MAXRANKNAME_Swiftmend,     MAXRANK_Swiftmend,     BOOKINDEX_Swiftmend      = Druid.maxRank("Swiftmend");
   end;

   -- -------------------------------------------------------------------------------------
   -- This function is called automatically by WoW when the interface first loads
   -- -------------------------------------------------------------------------------------

   onLoad = function()
   
      -----------------------------------------------------------------------------------   
      -- Since finding maxRank is a somewhat consuming process we'll find all the max ranks
      -- when the interface first loads.   Then we're set. If a skill is trained the
      -- max ranks are recalculated thanks to the event handler.
      -----------------------------------------------------------------------------------   

      Druid.SetAllMaxRanks()

      -----------------------------------------------------------------------------------   
      --Set up initial variables
      -----------------------------------------------------------------------------------   
      DFUNC_assist      = "off";           -- Initial setting for assisting a target
      assistName        = nil;             -- The name of the person to assist.
      assistType        = "caster"         -- Determine assist targeting type.
      DFUNC_Casting     = false            -- Set to true by the onEvent funciton during spellcast
            
      -----------------------------------------------------------------------------------   
      --Set up an event handler & slash command handler
      --Registered events are processed in DonEvent (set up in the xml file)
      -----------------------------------------------------------------------------------   
      this:RegisterEvent("UI_ERROR_MESSAGE")   
      this:RegisterEvent("SPELLS_CHANGED")
      this:RegisterEvent("SPELLCAST_START");
      this:RegisterEvent("SPELLCAST_STOP");
      this:RegisterEvent("SPELLCAST_INTERRUPTED");
      this:RegisterEvent("SPELLCAST_FAILED");
      this:RegisterEvent("PLAYER_REGEN_ENABLED");
      this:RegisterEvent("UNIT_INVENTORY_CHANGED");      
      this:RegisterEvent("PLAYER_ENTERING_WORLD");
      this:RegisterEvent("PLAYER_AURAS_CHANGED");
      this:RegisterEvent("PLAYER_UNGHOST");
      this:RegisterEvent("PLAYER_ALIVE");

      SLASH_DDruidFunc1 = "/Dassist";
      SlashCmdList["DDruidFunc"] = Druid.HandleSlashes;
      
      SLASH_DPlusHeal1   = "/DPlusHeal";
      SlashCmdList["DPlusHeal"] = Druid.SetPlusHeal;
      
      SLASH_DHardShift1 = "/DHardShift";
      SlashCmdList["DHardShift"] = Druid.SetHardShift;

      SLASH_DManaC1 = "/DManaC";
      SlashCmdList["DManaC"] = Druid.SetManaC;
    
      SLASH_DTarget1 = "/DTarget";
      SlashCmdList["DTarget"] = Druid.ShiftTarget;
      
      SLASH_DOutfit1 = "/DOutfit";
      SlashCmdList["DOutfit"] = Druid.Outfit;

      SLASH_DTrinkets1 = "/DTrinkets"
      SlashCmdList["DTrinkets"] = Druid.Trinkets;
     
      -----------------------------------------------------------------------------------   
      -- Hello World!
      -----------------------------------------------------------------------------------   
      DEFAULT_CHAT_FRAME:AddMessage( "DruidFunc Library Version 2.12b loaded and available");
   end;

   -- -------------------------------------------------------------------------------------
   -- This function is called whenever WoW finds an event we asked for in the onload area.
   -- This function is actually registered in the .toc file
   -- -------------------------------------------------------------------------------------

   DonEvent = function (event, reason)
       if (event == "UI_ERROR_MESSAGE") then
         if string.find(arg1, 'Cannot use while swimming') then
            Druid.ChangeShapeshiftForm("Aquatic Form")
         end
         if string.find(arg1, 'Can only use while swimming') then
            Druid.ChangeShapeshiftForm("Travel Form")
         end
      end
      if (event == "SPELLS_CHANGED") then
         -- Player learned a new spell, re-check all the max ranks used by this mod.
         Druid.SetAllMaxRanks()
      end;
      if (event == "SPELLCAST_START") then
         -- Keeps track of when we start a spell so we don't spam the chat box if user is mashing the keys.
         -- Also do a low mana check.
         DFUNC_Casting=true;
         if DFUNCSwap and UnitAffectingCombat("player") then
            if (Druid.GetCurrentForm() == 0) then
               Druid.LowManaSwap(85 );
            end
         end;
      end
      if (event == "UNIT_INVENTORY_CHANGED") then
         -- Player added or removed an inventory item, rescan the inventory.
         Druid.ScanInventory();
      end
      if (event == "SPELLCAST_STOP") then
         -- Player stopped casting the spell
         DFUNC_Casting=false;
      end
      if (event == "SPELLCAST_INTERRUPTED") then
         -- Player stopped casting the spell
         DFUNC_Casting=false;
      end
      if (event == "SPELLCAST_FAILED") then
         -- Player stopped casting the spell
         DFUNC_Casting=false;
      end
      if (event == "PLAYER_UNGHOST") or (event == "PLAYER_ALIVE") then
         -- Player recently rezzed or zoned, adjust his wardrobe
         if (DFUNCSwap) and not(UnitIsDeadOrGhost("Player")) then
            Druid.LoadWardrobe(DFUNCHumanSet)               
         end
      end
      if (event == "PLAYER_ENTERING_WORLD") then
         -- The game begins
         Druid.ScanInventory();
         DFUNCIsMounted = false;
      end
      if (event == "PLAYER_AURAS_CHANGED") then
         -- A buff was added or removed here we check for healing trinkets and if player mounted/dismounted
         
         DFUNCTrinketHeal = 0
         
         if Druid.isUnitBuffUp ("player", "Spell_Lightning_LightningBolt0") then
            DFUNCTrinketHeal = DFUNCTrinketHeal + 400
         end
         if Druid.isUnitBuffUp ("player", "Spell_Nature_HealingWaveGreater") then
            DFUNCTrinketHeal = DFUNCTrinketHeal + 350
         end
         
         if DFUNCIsMounted ~= Druid.UnitMounted() then
            DFUNCIsMounted = Druid.UnitMounted()
            if DFUNCSwap then
               if DFUNCIsMounted then
                  Druid.LoadWardrobe(DFUNCMountSet)
               else
                  Druid.LoadWardrobe(DFUNCHumanSet)               
               end
            end
         else
            Druid.WeaponShift();
         end
      end
      if (event == "PLAYER_REGEN_ENABLED") then
         if DFUNCSwap then
            if (Druid.GetCurrentForm() == 0) and not(Druid.isUnitBuffUp("player","Ability_Mount")) then
               Druid.LoadWardrobe(DFUNCHumanSet);
            end
         end;
         DFUNC_Casting = false;
      end;

   end;

-- -------------------------------------------------------------------------------------
-- testTrinket
--
-- This function looks for the specified string (IE "Zandalarian Hero Charm")
-- If the specified trinket is equipped and it's not in cooldown
-- this function will use that trinket.
--
-- Example  Druid.testTrinket("Zandalarian Hero Charm")
-- 
-- It's important that any trinket useage be done BEFORE you attempt to cast a spell.
-- A spell starts the global cooldown, using trinkets does not.  Therefore you can
-- use both trinkets and then cast a spell, but if you cast a spell first this function
-- will fail because the trinkets will be locked in the global cooldown.
--
-- Sample Macro usage
-- /script Druid.testTrinket("Zandalarian Hero Charm")
-- /script Druid.Moonfire()
--
-- This macro would automatically attempt to use any equiped hero charms every time you use moonfire.
--
-- Trinket code for the Zandalarin Hero charm and Mar'li's Eye are included (but commented out) 
-- in this lua.  If you're reading this file you probably have the skills to restore the functionality.
-- -------------------------------------------------------------------------------------


   testTrinket = function(lookfor) 
      local trinket1 = GetInventoryItemLink("player",GetInventorySlotInfo("Trinket0Slot"))
      local trinket2 = GetInventoryItemLink("player",GetInventorySlotInfo("Trinket1Slot"))

      lookfor = string.lower(lookfor);
      --Check Trinket Slot 0
      if (trinket1 ~= nil) then
         local _, _, itemCode = strfind(trinket1, "(%d+):");
         local itemName, _, _, _, _, itemType = GetItemInfo(itemCode);
         itemName = string.lower(itemName)
         if (itemName == lookfor) then
           if GetInventoryItemCooldown("player",13)==0 then
               UseInventoryItem( GetInventorySlotInfo("Trinket0Slot") );
           end; -- { check cooldown }
         end; -- { check trinket0 }
      end; -- {trinket1 not nil}

      --Check Trinket Slot 1
      if (trinket2 ~= nil) then
         local _, _, itemCode2 = strfind(trinket2, "(%d+):");
         local itemName2, _, _, _, _, itemType2 = GetItemInfo(itemCode2);
         itemName2 = string.lower(itemName2)
         if (itemName2 == lookfor) then
            if GetInventoryItemCooldown("player",14)==0 then
               UseInventoryItem( GetInventorySlotInfo("Trinket1Slot") );
            end; -- { check cooldown }
         end; -- {Check Trinket1 }
      end; -- {trinket2 not nil}
   end; -- {testTrinket}


-- -------------------------------------------------------------------------------------
-- This function is called whenver the player starts or changes equipment.
-- It scans all the items the player is wearing and calcuates the +heal value
-- -------------------------------------------------------------------------------------

   FindPattern = function (text, pattern, start)
      return string.sub(text, string.find(text, pattern, start))
   end;

   ScanInventory = function()
      if DFUNCNoScan then
         return
      end
      DFUNC_HealOffset = 0;
      ManaSet = false;
      HealSet = false;
      DMana5  = 0;
      DFunc_Tooltip:SetOwner(UIParent, "ANCHOR_NONE");
      for i, slotName in DFUNC_Slots do
         slotid, _ = GetInventorySlotInfo(slotName.. "Slot");
         foundItem = DFunc_Tooltip:SetInventoryItem("player", slotid);
         if ( foundItem ) then
            local numLines = DFunc_Tooltip:NumLines();
            for i=1, numLines, 1 do
               tmp = getglobal("DFunc_TooltipTextLeft"..i);
               tmpLine = tmp:GetText();
               tmpLine = string.lower(tmpLine);
               if tmpLine then
                  if string.find(tmpLine,"idol of rejuvenation") or string.find(tmpLine, "hibernation crystal") or string.find(tmpLine, "zandalarian hero") then
                     break
                  end
                  if string.find(tmpLine,"mana per") then
                     aString = Druid.FindPattern(tmpLine, "%d+");
                     if string.find(tmpLine,"set:") then
                        if not ManaSet then
                           ManaSet = true;
                        else
                           aString = 0
                        end
                     end
                     DMana5 = DMana5 + aString;
                  end
                  if string.find(tmpLine,"healing done by")  then
                     if not(string.find(tmpLine,"rejuvenation")) then
                        aString = Druid.FindPattern( tmpLine, "up to %d+")
                        aString = string.sub(aString,7,100);
                        if string.find(tmpLine,"set:") then
                           if not HealSet then
                              HealSet = true;
                           else
                              aString = 0
                           end
                        end
                        DFUNC_HealOffset = DFUNC_HealOffset + aString;
                     end
                  end
                  if string.find(tmpLine,"healing spells %+")  then
                     if not(string.find(tmpLine,"rejuvenation")) then
                        aString = Druid.FindPattern( tmpLine, "healing spells %+%d+")
                        aString = Druid.FindPattern( aString, "%d+")
                        if string.find(tmpLine,"set:") then
                           if not HealSet then
                              HealSet = true;
                           else
                              aString = 0
                           end
                        end
                        DFUNC_HealOffset = DFUNC_HealOffset + aString;
                     end
                  end
               end
            end;
         end
      end
   end;

-- -------------------------------------------------------------------------------------
-- MaxRank -- Attempt to discover the maximum rank of a spell
-- Credit to Adora -- http://www.gerdofal.net/wow/druid_mod.html
-- -------------------------------------------------------------------------------------

   FindActionButton = function ( spell )
      -- This procedure tries to find out where that spell is on the action bar
      -- once we locate the spell on the action bar we get all kinds
      -- of very nifty range and usability information about that spell.
      -- if the user doesn't have the spell on his toolbars then we return zero :(
      -- This procedure is heavilly modified from decursive so all credit and props where it is due.
      DFunc_Tooltip:SetOwner(UIParent, "ANCHOR_NONE");
      local i = 0;
      for i = 1, 120 do
         if (HasAction(i)) then
            if (GetActionText(i) == nil) then
               DFunc_Tooltip:ClearLines();
               DFunc_Tooltip:SetAction(i);
               local slotName = DFunc_TooltipTextLeft1:GetText();
               if (spell == slotName) then
                  return i;
               end
            end
         end
      end
      return 0;
   end;


   maxRank = function(spell)
      local i = 1;
      local maxrank = "rank 0";
      local castname;
      local found = false;
      bookindex = -1
      maxranknum = 0
      
      while true do
         local spellName, spellRank = GetSpellName(i,BOOKTYPE_SPELL)
         if not spellName then
            do break end
         end
         if (spellName == spell) then
            maxrank = spellRank;
            bookindex = i
            found = true;
         end
         i=i+1;
      end

      if not found then
         return maxrank, -1, -1;
      end
      
      local a=string.find(maxrank," ");
      local b=string.len(maxrank);
      if (b == Nil) or (a == Nil) then
         maxranknum = 0
      else
         maxranknum = tonumber(string.sub(maxrank,a,b));
      end

      return maxrank,maxranknum, bookindex;
   end;   
   

-- -------------------------------------------------------------------------------------
-- isUnitBuffUp (looks for the designated buff and returns true if found, false if not)
-- Credit to WowWiki
-- -------------------------------------------------------------------------------------

   dump = function () 
     local iIterator = 1
     while (UnitDebuff("target", iIterator)) do
       DEFAULT_CHAT_FRAME:AddMessage( "-->".. UnitDebuff("target", iIterator) );
       iIterator = iIterator + 1
     end
     return false
   end;

   isUnitBuffUp = function (sUnitname, sBuffname) 
     local iIterator = 1
     while (UnitBuff(sUnitname, iIterator)) do
       if (string.find(UnitBuff(sUnitname, iIterator), sBuffname)) then
          return true
       end
       iIterator = iIterator + 1
     end
     return false
   end;

   isUnitDebuffUp = function (sBuffname) 
     local iIterator = 1
     while (UnitDebuff("target", iIterator)) do
       if (string.find(UnitDebuff("target", iIterator), sBuffname)) then
          return true
       end
       iIterator = iIterator + 1
     end
     return false
   end;

  UnitMounted = function ()
     local i = 1
     Mounted = false
     while (UnitBuff("player", i)) do
        if (string.find(UnitBuff("player", i), "Ability_Mount")) then
           if not(string.find(UnitBuff("Player", i), "JungleTiger")) then
              Mounted = true
           end
        end
        i = i + 1
     end
     return Mounted
  end;


-- -------------------------------------------------------------------------------------
-- Mark of the wild
--
-- This function attempts to cast the highest possible rank of mark of the wild
-- on the selected target.
--
-- Sample Macro
-- /script Druid.MarkOfTheWild()
-- -------------------------------------------------------------------------------------

   MarkOfTheWild = function()
      local level = { 1, 10, 20, 30, 40, 50, 60 };
      local friend = UnitExists("target") and not UnitCanAttack("player", "target"); 
      local t = UnitLevel( "target" );
      local rank = MAXRANK_MoTW;
      
      if not friend then
         t = UnitLevel( "player" );
      end

      for i = rank, 1, -1 do
         if( t >= level[i] - 10 ) then
            CastSpellByName( "Mark of the Wild(Rank "..i..")" );
            if not friend then
               TargetUnit( "player" );
            end; -- {not friend}
            break;
         end; -- { target >= level }
      end; -- { for }
   end; -- {End MarkOfTheWild}

-- -------------------------------------------------------------------------------------
-- Thorns
-- -------------------------------------------------------------------------------------

   Thorns = function()
      local l = { 6, 14, 24, 34, 44, 54, 54, 54 };
      local t = UnitLevel( "target" );
      local friend = UnitExists("target") and not UnitCanAttack("player", "target"); 
      local rank = MAXRANK_Thorns;
      
      if not friend then
         t = UnitLevel( "player" );
      end; -- { not friend }

      for i = rank, 1, -1 do
         if( t >= l[i] - 10 ) then
            CastSpellByName( "Thorns(Rank "..i..")" );

            if not friend then
               TargetUnit( "player" );
            end; -- { not friend }
            
            break;
         end; -- { target > level }
      end; -- { for }
   end; -- { Thorns }

-- -------------------------------------------------------------------------------------
-- Abolish Poison
-- -------------------------------------------------------------------------------------

   AbolishPoison = function()
      local friend = UnitExists("target") and not UnitCanAttack("player", "target"); 
	
      CastSpellByName( "Abolish Poison" );
      
      if not friend then
         TargetUnit( "player" );
      end; -- {not friend}
   
   end; -- {Abolish Poison}

-- -------------------------------------------------------------------------------------
-- Remove Curse
-- -------------------------------------------------------------------------------------

   RemoveCurse = function()
      local friend = UnitExists("target") and not UnitCanAttack("player", "target"); 

      CastSpellByName( "Remove Curse" );

      if not friend then
         TargetUnit( "player" );
      end; -- { not friend }

   end; -- {Remove Curse}

-- -------------------------------------------------------------------------------------
-- Regrowth
-- -------------------------------------------------------------------------------------

   Regrowth = function( sendChat, targetEnemy )
      local l = { 12, 18, 24, 30, 36, 42, 48, 54, 60 };
      local t = UnitLevel( "target" );
      local friend = UnitExists("target") and not UnitCanAttack("player", "target"); 
      local h = UnitHealthMax( "target" ) - UnitHealth( "target" );
      local name = UnitName( "target" );
      if name == Nil then
         name = " "
      end

      local rank = MAXRANK_Regrow;

      if not friend then
         t = UnitLevel( "player" );
      end; -- { not friend }


      if DFUNCTrinkets then
         local pmana = UnitManaMax( "player" ) - UnitMana( "player" );
         if (pmana > 600) then
            Druid.testTrinket("Mar'li's Eye");
         end;
         Druid.testTrinket("Hibernation Crystal")         
      end;
      
      for i = rank, 1, -1 do
         if( t >= l[i] - 10 ) or (name == " ") then


            if ( h > 2000 ) and DFUNCTrinkets then
               Druid.testTrinket("Zandalarian Hero Charm");
            end; -- { h > 2000 }

            CastSpellByName( "Regrowth(Rank "..i..")" );

            if not friend and not UnitInRaid("player") then
               TargetUnit( "player" );
               name = UnitName( "player" );
            end;
            if sendChat and not DFUNC_Casting then
               DEFAULT_CHAT_FRAME:AddMessage( "Regrowth on "..name..". (Regrowth rank "..i.." )", "party" );
            end

            break;
         end; -- { target > level }
      end; -- { for }
      
      if targetEnemy then
         TargetLastEnemy();
      end; -- {targetEnemy}
      
   end; -- { Regrowth }

-- -------------------------------------------------------------------------------------
-- Rejuvenation -- Modified function.   This rejuvenation is extended to work with swiftmend
-- -------------------------------------------------------------------------------------

   Rejuvenation = function( sendChat, targetEnemy )
      local level = { 4, 10, 16, 22, 28, 34, 40, 46, 52, 58, 60 }; 
      local friend = UnitExists("target") and not UnitCanAttack("player", "target"); 
      local setTarget = "target";
      local retarget = false;

      if (not friend)  and not UnitInRaid("player") then
         setTarget = "player";
         retarget = true;
      end;

      local name        = UnitName( setTarget );        -- Name of the target
      if name == Nil then
         name = " "
      end
      TargetUnit(setTarget)
      
      if (Druid.isUnitBuffUp( setTarget, "Spell_Nature_Rejuvenation")) then
         if name ~= DFUNCRejuvWarn then
            if (BOOKINDEX_Swiftmend >= 0) then
               DEFAULT_CHAT_FRAME:AddMessage( "Rejuvenation already on target, press again to swiftmend!", 1.0, 0.0, 0.0, 1.0, 5  );
               PlaySound("Deathbind Sound")
            else
               DEFAULT_CHAT_FRAME:AddMessage( "Warning: Rejuvenation already on target, press again to refresh.", 1.0, 0.0, 0.0, 1.0, 5  );
               PlaySound("Deathbind Sound")
            end
            DFUNCRejuvWarn = name
            if retarget then
               TargetLastTarget()
            end
            if DFUNCSwap and UnitAffectingCombat("player") then
               if (Druid.GetCurrentForm() == 0) then
                  Druid.LowManaSwap(85);
               end
            end;
            return
         end
         DFUNCRejuvWarn = ""
         if (BOOKINDEX_Swiftmend >=0) then
            CastSpellByName("Swiftmend")
            if SpellIsTargeting() then
               SpellTargetUnit("Player")
               retarget = true
            end
            if retarget then
               TargetLastTarget()
            end
            if DFUNCSwap and UnitAffectingCombat("player") then
               if (Druid.GetCurrentForm() == 0) then
                  Druid.LowManaSwap(85);
               end
            end;
         end
      end

      DFUNCRejuvWarn = ""

      local targetLevel = UnitLevel( setTarget );       -- Level of the target
      rank              = MAXRANK_Rejuv;                -- Maxrank Rejuvenation set in onLoad

      if DFUNCTrinkets then
         local pmana = UnitManaMax( "player" ) - UnitMana( "player" );
         if (pmana > 600) then
            Druid.testTrinket("Mar'li's Eye");
         end;
         Druid.testTrinket("Hibernation Crystal")
      end;

      for i = rank, 1, -1 do
         if( targetLevel >= level[i] - 10 ) or (name == " ") then
            if sendChat and not DFUNC_Casting then
               DEFAULT_CHAT_FRAME:AddMessage( "HoT on "..name..". (Rejuvenation Rank "..i..")" );
            end; -- {sendChat}

            CastSpellByName( "Rejuvenation(Rank "..i..")" );
            

            if retarget then
               TargetLastTarget()
            end; -- { not friend }
            
            break;
         end; -- { target > level }
      end; -- { for }

      if DFUNCSwap and UnitAffectingCombat("player") then
         if (Druid.GetCurrentForm() == 0) then
            Druid.LowManaSwap(85);
         end
      end;


   end; -- { Rejuvenation }

-- -------------------------------------------------------------------------------------
-- Rejuvenation -- Original function.  This is the original, unrestricted rejuvenation
-- -------------------------------------------------------------------------------------

   Rejuv = function( sendChat, targetEnemy )
      local level = { 4, 10, 16, 22, 28, 34, 40, 46, 52, 58, 60 }; 
      local friend = UnitExists("target") and not UnitCanAttack("player", "target"); 
      local setTarget = "target";
      local retarget = false;

      if (not friend)  and not UnitInRaid("player") then
         setTarget = "player";
         retarget = true;
      end;

      local name        = UnitName( setTarget );        -- Name of the target
      TargetUnit(setTarget)
      
      local targetLevel = UnitLevel( setTarget );       -- Level of the target
      rank              = MAXRANK_Rejuv;                -- Maxrank Rejuvenation set in onLoad

      if DFUNCTrinkets then
         local pmana = UnitManaMax( "player" ) - UnitMana( "player" );
         if (pmana > 600) then
            Druid.testTrinket("Mar'li's Eye");
         end;
         Druid.testTrinket("Hibernation Crystal")
      end;

      for i = rank, 1, -1 do
         if( targetLevel >= level[i] - 10 ) or (name == " ") then
            if sendChat and not DFUNC_Casting then
               DEFAULT_CHAT_FRAME:AddMessage( "HoT on "..name..". (Rejuvenation Rank "..i..")" );
            end; -- {sendChat}

            CastSpellByName( "Rejuvenation(Rank "..i..")" );
            

            if retarget then
               TargetLastTarget()
            end; -- { not friend }
            
            break;
         end; -- { target > level }
      end; -- { for }

      if DFUNCSwap and UnitAffectingCombat("player") then
         if (Druid.GetCurrentForm() == 0) then
            Druid.LowManaSwap(85);
         end
      end;

   end; -- { Rejuv }

-- -------------------------------------------------------------------------------------
-- Healing Touch (trinket version)
-- PowerTouch exists soley to trip the zandalaar druid class trinket before calling
-- Healing touch.
-- -------------------------------------------------------------------------------------

   PowerTouch = function( minRank, sendChat, targetEnemy )
      Druid.testTrinket("Wushoolay's Charm of Nature");
      Druid.testTrinket("Zandalarian Hero Charm");
      Druid.HealingTouch( minRank, sendChat, targetEnemy )   
   end;

-- -------------------------------------------------------------------------------------
-- Healing Touch
-- -------------------------------------------------------------------------------------

   HealingTouch = function( minRank, sendChat, targetEnemy, forceRank )
      local friend = UnitExists("target") and not UnitCanAttack("player", "target"); 
      local party = GetNumPartyMembers() > 0;
      local setTarget = "target";

      rank=MAXRANK_HealingTouch;
     
      if (not friend) and not UnitInRaid("player") then
         setTarget = "player";
      end;

      local name = UnitName( setTarget );
      
      if name == Nil then
         name = " "
      end

      if DFUNCTrinkets then
         local pmana = UnitManaMax( "player" ) - UnitMana( "player" );
         if (pmana > 600) then
            Druid.testTrinket("Mar'li's Eye");
         end;
         Druid.testTrinket("Hibernation Crystal")
      end;

      if (not friend) or (UnitHealthMax( setTarget ) > 100) then
         local sr =  { 0,   58, 125, 266, 482, 748, 960,  1200, 1526, 1917, 2267 };
         local sr2 = { 58, 125, 266, 482, 748, 960, 1200, 1526, 1917, 2370, 2677 }; 
         local h = UnitHealthMax( setTarget ) - UnitHealth( setTarget );

         for i = rank, 1, -1 do
            if DFUNCPlusHeal == "auto" then
               offset = DFUNC_HealOffset
            else
               offset = DFUNCPlusHeal
            end
            if offset > 0 then 
               offset = offset + DFUNCTrinketHeal
            end
            healpower = sr[i] + offset;
            if ( (h >= healpower) or ( sr[i]==0 ) ) then
               if ( i < minRank ) then
                  i = minRank
               end
               if (forceRank) then
                  i = minRank;
               end
               if sendChat and not DFUNC_Casting then
                  DEFAULT_CHAT_FRAME:AddMessage( "Healing "..name..". (Rank "..i..") "..sr2[i]+offset.." heal/"..h.." needed", "party" );
               end 
                              
               if not DFUNC_Casting then
                  CastSpellByName( "Healing Touch(Rank "..i..")" );
               else
                  if (DFUNCManaC ~= "off") then
                     local percentHealth = floor(UnitHealth( setTarget ) / UnitHealthMax( setTarget ) * 100);
                     if ( percentHealth >= tonumber(DFUNCManaC) ) then
                        if not(DFUNCSilent) then
                           DEFAULT_CHAT_FRAME:AddMessage( "Mana Conserve aborted the heal. Target @ "..percentHealth.."% abort set to "..DFUNCManaC.."%", 1.0, 0.0, 0.0, 1.0, 5 );
                        end
                        SpellStopCasting()
                        DFUNC_Casting = false;
                     end
                  end
               end
               if not friend and not UnitInRaid("player") then
                  SpellTargetUnit( "player" );
               end
               if SpellIsTargeting() and not UnitInRaid("player") then
                  SpellStopCasting()
                  DEFAULT_CHAT_FRAME:AddMessage( "Heal aborted.  Invalid target.", 1.0, 0.0, 0.0, 1.0, 5 );
               end
                  
               break;
            end
         end
      else
         local level = { 0, 8, 14, 20, 26, 32, 38, 44, 50, 56, 60 }; 
         local t = UnitLevel( setTarget );
         local h1 = UnitHealthMax( setTarget );
         local h2 = UnitHealth( setTarget );
         local h = UnitHealthMax( setTarget ) - UnitHealth( setTarget );

         for i = rank, 1, -1 do
            if( t >= level[i] - 10 ) then
               if ( i < minRank ) then
                  i = minRank
               end
               if forceRank then
                  i = minRank;
               end
               if sendChat and not DFUNC_Casting then
                  DEFAULT_CHAT_FRAME:AddMessage( "Full Heal rank " ..i.. " (target health: ".. h ..", health max: ".. h1 .. ", curr Health: " .. h2 .. ")", "party" );
               end
               CastSpellByName( "Healing Touch(Rank "..i..")" );
               if not friend and not UnitInRaid("player") then
                  SpellTargetUnit( "player" );
               end
               if SpellIsTargeting() and not UnitInRaid("player") then
                  SpellStopCasting()
                  DEFAULT_CHAT_FRAME:AddMessage( "Heal aborted.  Invalid target.", 1.0, 0.0, 0.0, 1.0, 5 );
               end
               break;
            end
         end
      end
      if targetEnemy then
         TargetLastEnemy();
      end
   end;
-- -------------------------------------------------------------------------------------
-- Self Cast Innervate
-- -------------------------------------------------------------------------------------

   ManaMe = function ()
      TargetUnit("player")
      Druid.Innervate(true, false)
      TargetLastTarget()
   end;

-- -------------------------------------------------------------------------------------
-- Innervate
-- -------------------------------------------------------------------------------------

   Innervate = function( sendChat, targetEnemy )
      local name = UnitName( "target" );

      if name == Nil then 
         name = UnitName( "player" ) 
      end 
      
      local pname = UnitName("player")
      local friend = UnitExists("target") and not UnitCanAttack("player", "target"); 

      if UnitInRaid( "player" ) then
         chatTarget = "raid"
      else
         if ( GetNumPartyMembers() > 0 ) then
            chatTarget = "party"
         else
            chatTarget = "say"
         end
      end

      if Druid.isUnitBuffUp("target", "Nature_Lightning") then
         DEFAULT_CHAT_FRAME:AddMessage( "Innervate already on target.  Aborted", 1.0, 0.0, 0.0, 1.0, 5  );
         return
      end      
      
      if not(friend) then
         TargetUnit("player")
         local name = UnitName( "target" );
      else
         if (UnitPowerType("target") ~= 0) then
            TargetUnit("player")
            local name = UnitName( "target" );
         end
      end
      
      local pmana = floor(UnitMana( "target" ) / UnitManaMax( "Target" ) * 100);
      
      if pmana > 50 then
         DEFAULT_CHAT_FRAME:AddMessage( "Innervate aborted.  "..name.." has more than 50% mana", 1.0, 0.0, 0.0, 1.0, 5 );
         name2 = UnitName("target")
         if name ~= name2 then
            TargetLastTarget()
         end
         return
      end

      DEFAULT_CHAT_FRAME:AddMessage( "Casting Innervate on "..name );

      CastSpellByName("Innervate")

      local name2 = UnitName("target")
      if name2 ~= pname then
         if (party and sendChat) then      
            SendChatMessage("Casting Innervate on "..name, chatTarget );      
         end
      else
         DEFAULT_CHAT_FRAME:AddMessage( "Casting Innervate on "..name );
         if DFUNCSwap then
            Druid.LoadWardrobe(DFUNCISet)
         end
      end
      if name ~= name2 then
         TargetLastTarget()
      end
      
   end;


-- -------------------------------------------------------------------------------------
-- Moonfire
-- -------------------------------------------------------------------------------------

   Moonfire = function()
      if BOOKINDEX_Moonfire < 0 then
         BOOKINDEX_Moonfire = Druid.FindActionButton("Moonfire");
      end;
      Druid.HandleAssist( BOOKINDEX_Moonfire );
      local rank = MAXRANK_Moonfire;
      local name = UnitName( "target" );
      if DFUNCTrinkets then
         Druid.testTrinket("Zandalarian Hero Charm");
      end
      if( UnitName( "target" ) ~= nil ) then
         if ( string.find( name, "Totem" ) ~= nil ) and  ( string.find( name, "Totemic" ) == nil ) then
            CastSpellByName( "Moonfire(Rank 1)" );
         else
            CastSpellByName( "Moonfire(Rank "..rank..")" );
         end
      end
      Druid.CleanupAssist();
   end;

-- -------------------------------------------------------------------------------------
-- Starfire
-- -------------------------------------------------------------------------------------

   Starfire = function()
      if BOOKINDEX_Starfire < 0 then
         BOOKINDEX_Starfire = Druid.FindActionButton("Starfire");
      end;
      Druid.HandleAssist( BOOKINDEX_Starfire );
      local rank = MAXRANK_Starfire;
      if DFUNCTrinkets then
         Druid.testTrinket("Zandalarian Hero Charm");
      end
      CastSpellByName( "Starfire(Rank "..rank..")" );
      Druid.CleanupAssist();
   end;

-- -------------------------------------------------------------------------------------
-- Wrath
-- -------------------------------------------------------------------------------------
 
   Wrath = function()
      if BOOKINDEX_Wrath < 0 then
         BOOKINDEX_Wrath = Druid.FindActionButton("Wrath");
      end;
      Druid.HandleAssist( BOOKINDEX_Wrath );

      local rank = MAXRANK_Wrath;

      
      if DFUNCTrinkets then
         Druid.testTrinket("Zandalarian Hero Charm");
      end
      CastSpellByName( "Wrath(Rank "..rank..")" );

      Druid.CleanupAssist();
   end;

-- -------------------------------------------------------------------------------------
-- Insect Swarm
-- -------------------------------------------------------------------------------------

   InsectSwarm = function()
      if BOOKINDEX_InsectSwarm < 0 then
         BOOKINDEX_InsectSwarm = Druid.FindActionButton("Insect Swarm");
      end;
      Druid.HandleAssist( BOOKINDEX_InsectSwarm );
      local rank = MAXRANK_InsectSwarm;
      if DFUNCTrinkets then
         Druid.testTrinket("Zandalarian Hero Charm");
      end
      CastSpellByName( "Insect Swarm(Rank "..rank..")" );      
      Druid.CleanupAssist();
   end;

-- -------------------------------------------------------------------------------------
-- Faerie Fire
-- -------------------------------------------------------------------------------------

   FaerieFire = function()
      if BOOKINDEX_FF < 0 then
         BOOKINDEX_FF = Druid.FindActionButton("Faerie Fire");
      end;
      Druid.HandleAssist( BOOKINDEX_FF );
      local rank = MAXRANK_FF;
      if (Druid.GetCurrentForm() == 0) then
         CastSpellByName( "Faerie Fire(Rank "..rank..")" );      
      else
         CastSpellByName( "Faerie Fire (Feral)(Rank "..MAXRANK_FFF..")" );
      end;
      Druid.CleanupAssist();
   end;

-- -------------------------------------------------------------------------------------
-- Barkskin
-- -------------------------------------------------------------------------------------
	
   Barkskin = function ()
      if DFUNCTrinkets then
         Druid.testTrinket("Defiler's Talisman");
      end
      CastSpellByName( "Barkskin" );
   end;

-- -------------------------------------------------------------------------------------
-- Nature's Grasp
-- -------------------------------------------------------------------------------------
	
   NaturesGrasp = function ()
      CastSpellByName( "Nature's Grasp(Rank "..MAXRANK_NaturesGrasp..")" );
   end;

-- -------------------------------------------------------------------------------------
-- Hurricane
-- -------------------------------------------------------------------------------------
	
   Hurricane = function ()
      CastSpellByName( "Hurricane(Rank "..MAXRANK_Hurricane..")" );
   end;
   
-- -------------------------------------------------------------------------------------
-- Entangling Roots
-- -------------------------------------------------------------------------------------
	
   Roots = function ()
      CastSpellByName( "Entangling Roots(Rank "..MAXRANK_Roots..")" );
   end;
   
-- -------------------------------------------------------------------------------------
-- Tranquility
-- -------------------------------------------------------------------------------------
	
   Tranquility = function ()
      CastSpellByName( "Tranquility(Rank "..MAXRANK_Tranquility..")" );
   end;
   
-- -------------------------------------------------------------------------------------
-- Hibernate
-- if partyChat is true this will inform the raid/party that you have hibernated a target
-- This function is not bound to the auto-assist routies.
-- -------------------------------------------------------------------------------------
   
   Hibernate = function ( partyChat ) 

      if not DFUNC_Casting then      
         if UnitCreatureType("target")=="Beast" or UnitCreatureType("target")=="Dragonkin" then
            CastSpellByName( "Hibernate(Rank "..MAXRANK_Hibernate..")" )
            if not Druid.isUnitDebuffUp("Spell_Nature_Sleep") then
               if partyChat then
                  local name = UnitName( "target" );
                  if UnitInRaid( "player" ) then
                    chatTarget = "raid"
                  else
                     if ( GetNumPartyMembers() > 0 ) then
                        chatTarget = "party"
                     else
                        chatTarget = "say"
                     end
                  end
                  SendChatMessage("Casting hibernate (CC) on "..name.." <DO NOT ATTACK!>", chatTarget );      
               end
            end
         else
               DEFAULT_CHAT_FRAME:AddMessage( "Invalid hibernation target.");
         end
      end      
   end;

-- -------------------------------------------------------------------------------------
-- Rebirth
-- if partyChat is true this will inform the raid/party that you have hibernated a target
-- This function is not bound to the auto-assist routies.
-- -------------------------------------------------------------------------------------
   
   Rebirth = function ( partyChat )
      local name   = UnitName( "target" )
      local friend = UnitExists("target") and not UnitCanAttack("player", "target");
      local ghost  = UnitIsGhost( "target" )

      if not DFUNC_Casting then
         if friend and not ghost then
            CastSpellByName( "Rebirth(Rank "..MAXRANK_Rebirth..")" )
            if partyChat then
               if UnitInRaid( "player" ) then
                 chatTarget = "raid"
               else
                  if ( GetNumPartyMembers() > 0 ) then
                     chatTarget = "party"
                  else
                     chatTarget = "say"
                  end
               end
               SendChatMessage("Resurrecting "..name, chatTarget );      
            end
         else
            DEFAULT_CHAT_FRAME:AddMessage( "Invalid Rebirth target.");
         end
      end
      
   end;

-- -------------------------------------------------------------------------------------
-- Desperate Prayer (Automatically casts a maximum heal on the player -- with nature's swiftness if available)
-- -------------------------------------------------------------------------------------

   DesperatePrayer = function ()
      local percentHealth = floor(UnitHealth( "player" ) / UnitHealthMax( "player" ) * 100);
      currentForm, name = Druid.GetCurrentForm()      
      if (currentForm ~= 0) then
         Druid.ShiftHuman();
      end
      SpellStopCasting();      
      TargetUnit("player")
      if percentHealth < 50 then
         CastSpellByName( "Nature's Swiftness" )
         SpellStopCasting();
      end
      CastSpellByName( "Healing Touch(Rank "..MAXRANK_HealingTouch..")");
      TargetLastTarget()
   end;

-- -------------------------------------------------------------------------------------
-- Feral Section
-- -------------------------------------------------------------------------------------

   Ravage = function ()
      CastSpellByName( "Ravage(Rank "..MAXRANK_Ravage..")" );
   end;

   Shred = function ()
      CastSpellByName( "Shred(Rank "..MAXRANK_Shred..")" );
   end;

   Claw = function ()
      CastSpellByName( "Claw(Rank "..MAXRANK_Claw..")" );
   end;
   
   Prowl = function ()
      CastSpellByName( "Prowl(Rank "..MAXRANK_Prowl..")" );
   end;

   Rake = function ()
      CastSpellByName( "Rake(Rank "..MAXRANK_Rake..")" );
   end;

   Rip = function ()
      CastSpellByName( "Rip(Rank "..MAXRANK_Rip..")" );
   end;

   Pounce = function ()
      CastSpellByName( "Pounce(Rank "..MAXRANK_Pounce..")" );
   end;
   
   FeralCharge = function ()
      CastSpellByName ( "Feral Charge" );
   end;

   FerociousBite = function ()
      CastSpellByName ( "Ferocious Bite(Rank "..MAXRANK_FerociousBite..")" );
   end;

   TigersFury = function ()
      CastSpellByName ( "Tiger's Fury(Rank "..MAXRANK_TigersFury..")" );
   end;

   Cower = function ()
      CastSpellByName ( "Cower(Rank "..MAXRANK_Cower..")" );
   end;

   Dash = function ()
      CastSpellByName ( "Dash(Rank "..MAXRANK_Dash..")" );
   end;

   Maul = function ()
      CastSpellByName( "Maul(Rank "..MAXRANK_Maul..")" );
   end;

   Swipe = function ()
      CastSpellByName( "Swipe(Rank "..MAXRANK_Swipe..")" );
   end;

   Bash = function ()
      CastSpellByName( "Bash(Rank "..MAXRANK_Bash..")" );
   end;

   Growl = function ()
      CastSpellByName( "Growl" );
   end;

   Enrage = function ()
      CastSpellByName( "Enrage" );
   end;

   ChalRoar = function ()
      CastSpellByName( "Challenging Roar" );
   end;

   FrenziedRegen = function ()
      CastSpellByName( "Frenzied Regeneration(Rank "..MAXRANK_FrenziedRegen..")" );
   end;

   DemRoar = function ()
      CastSpellByName( "Demoralizing Roar(Rank "..MAXRANK_DemRoar..")" );
   end;
   


   BestAttackCat = function ()
      if UnitAffectingCombat( "player" ) then
         if (GetComboPoints() >= 5) then
            Druid.FerociousBite();
         else
            if UnitMana("player") > 35 then
               Druid.Claw();
            else
               Druid.FaerieFire()
            end
         end
      else
         if Druid.isUnitBuffUp("player", "Ability_Ambush") then
            if UnitMana("player") > 90 then
               Druid.TigersFury();
               SpellStopCasting();
            end
            Druid.Ravage();
         else
            CastSpellByName("Track Humanoids");
            SpellStopCasting()
            Druid.Prowl();
         end
      end;
   end;

-- -------------------------------------------------------------------------------------
-- Swiftshift section -- This area holds functions and procedures which handle shapeshifting
-- -------------------------------------------------------------------------------------

   GetCurrentForm = function ()
      local maxForms = GetNumShapeshiftForms();
      local i=1;
      local currentForm = 0;
      local currentName = "";
      for i=1, maxForms, 1 do
        icon, name, active, castable = GetShapeshiftFormInfo(i);
        if (active ~= nil) then
           currentForm = i
           currentName = name
        end
      end
      if currentForm == 0 then 
         currentName = "Humanoid Form"
      end
      return currentForm, currentName;
   end;

   WeaponShift = function()
      -- Called when the user changes forms
      
      if not Druid.isUnitBuffUp("Player", "Nature_Lightning") and (DFUNCSwap) then
         currentForm, name = Druid.GetCurrentForm()
         name=string.lower(name);
         if name ~= DFUNCLastForm then
            DFUNCLastForm = name;
            if string.find(name, "bear") then
               if DFUNCBearMain ~= "off" then
                  Druid.LoadWardrobe(DFUNCBearSet);
               end
            end
            if string.find(name, "cat") then
               if DFUNCCatMain ~= "off" then
                  Druid.LoadWardrobe(DFUNCCatSet)
               end
            end
            if string.find(name, "humanoid") then
               if (DFUNCHumanMain ~= "off") then
                  if not(UnitAffectingCombat("player")) then
                     Druid.LoadWardrobe(DFUNCHumanSet)
                  end
               end
            end
            if string.find(name, "moonkin") then
               if DFUNCMoonMain ~= "off" then
                  Druid.LoadWardrobe(DFUNCMoonSet);
               end
            end
            if string.find(name, "travel") then
               if DFUNCHumanMain ~= "off" then
                  Druid.LoadWardrobe(DFUNCTravelSet);
               end
            end
            if string.find(name, "aqua") then
               if DFUNCAquaMain ~= "off" then
                  Druid.LoadWardrobe(DFUNCAquaSet);
               end
            end
         end
      end
   end;

   ChangeShapeshiftForm = function ( shift )
	
	local i = 1
	local target =0;
	local current = 0;
	local maxForms = GetNumShapeshiftForms();
	
	if UnitClass("Player") ~= "Druid" then
	   return
	end

        if Druid.isUnitBuffUp("player","Ability_Mount") then
           local i = 0
           while (GetPlayerBuffTexture(i)) do
              if (string.find(GetPlayerBuffTexture(i), "Ability_Mount")) then
                 if not(string.find(GetPlayerBuffTexture(i), "JungleTiger")) then
                    CancelPlayerBuff(i);
                    SpellStopCasting();
                 end
              end
              i = i + 1
           end
        end; 
	
	if DFUNCTrinkets then
           Druid.testTrinket("Rune of Metamorphosis");
        end
        
        for i = 1, maxForms, 1 do
           icon, name, active, castable = GetShapeshiftFormInfo(i);
           if (active ~= nil) then
              current = i
           end
           if ( string.find( name, shift ) ~= nil ) then
              target = i
           end
        end;
        if (target==0) and (current==0) then
           --User wants humanoid and is already humanoid
           --So target the user. This allows F1 funcationality to be maintained 
           --If the user elects to use function keys for swiftshifting.
           if (DFUNCHardShift=="on") and (DFUNCSwiftTarget) then
              TargetUnit( "player" )
           else
              if (DFUNCHardShift=="on") then
                 local percentHealth = floor(UnitHealth( "player" ) / UnitHealthMax( "player" ) * 100);
                 local friend = UnitExists("target") and not UnitCanAttack("player", "target"); 
                 if friend then 
                    TargetUnit( "player" )
                 end         
                 if percentHealth < 25 then
                    CastSpellByName( "Nature's Swiftness" )
                    SpellStopCasting();
                    CastSpellByName( "Healing Touch(Rank "..MAXRANK_HealingTouch..")");
                    SpellTargetUnit( "player" )
                    TargetLastTarget();
                 else
                    CastSpellByName("Regrowth(Rank "..MAXRANK_Regrow..")");
                    SpellTargetUnit( "player" )
                    TargetLastTarget();
                 end
              else
                 TargetUnit( "player" )
              end;
           end
        else           
           if (current==0) then
              CastShapeshiftForm(target);
           else
              if (DFUNCHardShift=="on") and (shift ~="Travel Form") then
                 if (current ~= target) then
                    CastShapeshiftForm(current);
                 else
                    if (shift=="Cat Form") then
                       Druid.BestAttackCat()
                    end
                    if (shift=="Bear Form") then
                       Druid.FeralCharge();
                       Druid.Maul();
                    end
                 end
              else
                 if (target==0) then
                    target=current
                 end
                 CastShapeshiftForm(current);
              end
           end
        end;
   end;
   
   ShiftBear = function ()
      Druid.ChangeShapeshiftForm("Bear Form")
      
   end;
   
   ShiftCat = function ()
      Druid.ChangeShapeshiftForm("Cat Form")
   end;
   
   ShiftTravel = function ()
      Druid.ChangeShapeshiftForm("Travel Form")
   end;

   ShiftAqua = function ()
      Druid.ChangeShapeshiftForm("Aquatic Form")
   end;
   
   ShiftMoon = function ()
      Druid.ChangeShapeshiftForm("Moonkin Form")
      CURRENT_ACTIONBAR_PAGE = 5;
      ChangeActionBarPage();      
   end;

   ShiftHuman = function ()
      Druid.ChangeShapeshiftForm("Humanoid Form")
   end;
   
-- -------------------------------------------------------------------------------------
-- Assist section
--
-- These functions handle auto assist.  
-- If the assist is a ghost (dead) you'll attack whatever you have targeted.
--
-- It's recommended you activate this function only in raid dungeons
-- In PvP and dynamic outdoor situations setting an assist will get you killed more often than not.
--
-- Usage:
-- /dassist                         With no arguments this just shows who you are assisting
-- /dassist <name>                  This turns on the assist
-- /dassist off                     This turns off the assist
-- /dassist on                      This turns on the assist and uses the last assist name.
-- -------------------------------------------------------------------------------------
   
   HandleAssist = function (actionSlot)
      DFUNC_LastTarget = nil;

      if DFUNC_assist ~= "off" then
         DFUNC_LastTarget = UnitName( "target" );
         if (DFUNC_LastTarget == nil) then
            DFUNC_LastTarget = UnitName( "player" );
         end
         TargetByName(DFUNC_assist);
         if UnitAffectingCombat("target") then
            if not UnitIsDeadOrGhost( "target" ) then              
               AssistByName(DFUNC_assist);
               local friend = UnitExists("target") and not UnitCanAttack("player", "target");            
               if (not UnitIsVisible("target")) or friend then
                  TargetByName(DFUNC_LastTarget);
                  DEFAULT_CHAT_FRAME:AddMessage( "Assist Failed... Not Visible or doesn't exist or not hostile");
               else
                  if actionSlot > 0 then
               
                     if IsActionInRange(actionSlot) ~= 1 then
                        TargetByName(DFUNC_LastTarget);
                        DEFAULT_CHAT_FRAME:AddMessage( "Assist Failed... Out of Range");
                     end;

                  end;
               end;
            else
               TargetByName(DFUNC_LastTarget);
               DEFAULT_CHAT_FRAME:AddMessage( "Assist Failed... Assist is dead or bad target.");
            end
         else
            TargetByName(DFUNC_LastTarget);
            DEFAULT_CHAT_FRAME:AddMessage( "Assist failed...  The person you are assisting is not in combat.");
         end
      end
   end;
   
   CleanupAssist = function ()
      if (DFUNC_LastTarget ~= nil) and (DFUNC_assist ~= nil) and (assistType == "caster") then
         TargetByName(DFUNC_LastTarget);
      end
   end;
   
-- -------------------------------------------------------------------------------------
-- Slash Command Handlers
-- -------------------------------------------------------------------------------------

   SetPlusHeal = function (msg)
      -- This function is called by the event handler when the /dplusheal slash command is found
      if (string.lower(msg) == "off") then
         msg = "0"
      end
      if (tonumber(msg) == nil) then
         msg = string.lower(msg);
         if msg == "auto" then
            DFUNCPlusHeal = "auto"
            DEFAULT_CHAT_FRAME:AddMessage( "Healing touch will now be automatically offset by your +heal items.");
            DEFAULT_CHAT_FRAME:AddMessage( "Your present +heal value from items: "..DFUNC_HealOffset);
         else
            DEFAULT_CHAT_FRAME:AddMessage( "Healing touch is being offset by "..DFUNCPlusHeal);
            if (DFUNCPlusHeal == "auto") then
               DEFAULT_CHAT_FRAME:AddMessage( "Your present +heal value from items: "..DFUNC_HealOffset);
            end
         end;
         DEFAULT_CHAT_FRAME:AddMessage( "Your present mana/5 value from items: "..DMana5);
         local base, stat, posBuff, negBuff = UnitStat("player",5)         
         SpiritTick = floor(stat/5) + 13
         DEFAULT_CHAT_FRAME:AddMessage( "Your present spirit regen (outside the 5 sec rule): "..SpiritTick);
         Innervate = SpiritTick * 40
         DEFAULT_CHAT_FRAME:AddMessage( "Innervate will return around : "..Innervate.." mana (disregarding reflection and tier 2 set bonus)");
      else
         DFUNCPlusHeal = tonumber(msg)
         DEFAULT_CHAT_FRAME:AddMessage( "Healing touch is now offset by "..DFUNCPlusHeal..". This number is saved between sessions");
         DEFAULT_CHAT_FRAME:AddMessage( "Use /DPlusHeal 0 to stop using this feature.");
         DEFAULT_CHAT_FRAME:AddMessage( "/DPlusHeal with no number will display the current setting.");
      end;
   end;

   SetHardShift = function (msg)
      -- This function is called by the event handler when the /dHardShift slash command is found
      if msg=="on" then
         DFUNCHardShift = "on";
      end
      if msg == "off" then
         DFUNCHardShift = "off";
      end
      DEFAULT_CHAT_FRAME:AddMessage( "Hardshifting is presently: "..DFUNCHardShift);
   end;
   
   trim = function( aString )
      if aString ~= nil then
         aString = string.gsub( aString, "^[%c%s]+", "" )
         aString = string.gsub( aString, "[%s%c]+$", "" )
      end
      return aString
   end;
 
   Trinkets = function (msg)
      msg = string.lower(msg)
      if (msg == "on") then
         DFUNCTrinkets = true
      end
      if (msg == "off") then
         DFUNCTrinkets = false
      end
      if DFUNCTrinkets then
         DEFAULT_CHAT_FRAME:AddMessage( "Pre-Defined Auto Trinkets are enabled");
      else
         DEFAULT_CHAT_FRAME:AddMessage( "Pre-Defined Auto Trinekts are disabled");
      end
   end;
 
 
   SetManaC = function (msg)
      -- This function is called by the event handler when the /dManaC slash command is found
      msg = string.lower(msg)
      if (msg == "silent") then
         DFUNCSilent = true
         DEFAULT_CHAT_FRAME:AddMessage( "Mana Conserve will no longer announce aborts in chat");
         return
      end
      if (msg == "announce") then
         DFUNCSilent = false
         DEFAULT_CHAT_FRAME:AddMessage( "Mana Conserve announce aborts to you");
         return
      end
      
      if (msg == "off") then
         DFUNCManaC = "off";
      else
         x = tonumber(msg)
         if x==Nil then
            x=0
         end
         if (x >= 10) and (x <=100) then
            DFUNCManaC = x;
         end
      end
      if DFUNCManaC == "off" then
            DEFAULT_CHAT_FRAME:AddMessage( "Mana Conserve has been set to OFF, spells will not abort.");
      else
            DEFAULT_CHAT_FRAME:AddMessage( "Mana Conserve is ON and will abort spells if target has more than "..DFUNCManaC.."% health.");
      end
   end;

   ShiftTarget = function (msg)
      msg = string.lower(msg)
      if msg == "off" then
         DEFAULT_CHAT_FRAME:AddMessage( "Hardshifting will now cast self-heal on repeated humanoid keypresses");
         DFUNCSwiftTarget = false;
      else
         DEFAULT_CHAT_FRAME:AddMessage( "Repeated humanoid keypresses will now self-target the player");
         DFUNCSwiftTarget = true;
      end
   end;



   HandleSlashes = function (msg)
        if msg == nil then
           msg = " "
        end
	msg = string.lower(msg);
	
	if (msg=="") then
	   if DFUNC_assist=="off" then
              DEFAULT_CHAT_FRAME:AddMessage( "Auto-assist is currently disabled");
              if assistName ~= nil then
                 DEFAULT_CHAT_FRAME:AddMessage( "/dassist on will assist: "..assistName);
              end              
	   else
              DEFAULT_CHAT_FRAME:AddMessage( "Currently Assisting:"..DFUNC_assist);
           end
           DEFAULT_CHAT_FRAME:AddMessage( "Use /Dassist off to disable | Use /Dassist on to re-enable using last assist");
           DEFAULT_CHAT_FRAME:AddMessage( "Use /Dassist <name> to set an assist (and enable assisting)");
           DEFAULT_CHAT_FRAME:AddMessage( "Use /Dassist feral  -- changes your target to your assist's target.");           
           DEFAULT_CHAT_FRAME:AddMessage( "Use /Dassist caster -- Always returns to your target after an attack.");           
           DEFAULT_CHAT_FRAME:AddMessage( "If your assist is a ghost or not targeting an enemy you'll attack whatever you have selected.");
        else
           local a,b,cmd=strfind(msg, "(%S+)");
	   if (cmd=="off") then
	      DFUNC_assist="off";
              DEFAULT_CHAT_FRAME:AddMessage( "Assisting Disabled");
           else
              if (cmd=="on") then
                 if assistName ~= nil then
                    DFUNC_assist=assistName;
                    DEFAULT_CHAT_FRAME:AddMessage( "Assist set to: "..DFUNC_assist);
                 else
                    DEFAULT_CHAT_FRAME:AddMessage( "Failed!  No assist name has been set.  Use /dassist <name>");
                 end
               else
                  if (cmd=="feral") then
                     assistType="feral"
                     DEFAULT_CHAT_FRAME:AddMessage( "Feral assist enabled -- will always target your assist's target.");
                  else
                     if (cmd=="caster") then
                        assistType="caster"
                        DEFAULT_CHAT_FRAME:AddMessage( "Caster assist enabled -- will return to your current target after spellcast");
                     else
                        if cmd == nil then
                           cmd = "off"
                        end
                        DFUNC_assist=cmd;
                        assistName=cmd;
                        DEFAULT_CHAT_FRAME:AddMessage( "Assist set to: "..cmd);
                     end
                  end
               end
           end;
        end
	   
   end;
   
-- -------------------------------------------------------------------------------------
-- This section handles the wardrobe changes on shift
-- -------------------------------------------------------------------------------------

   ScanGump = function()
      -- -------------------------------------------------------------------------------------
      -- This function scans and records all the items the user is currently wearing
      -- and returns them in a table
      -- -------------------------------------------------------------------------------------
      DFunc_Tooltip:SetOwner(UIParent, "ANCHOR_NONE");
      ktr=0;
      tmpGump = {}
      for i, slotName in DFUNC_Slots do
         ktr = ktr + 1
         tmpGump[ktr] = "";
         slotid, _ = GetInventorySlotInfo(slotName.. "Slot");
         foundItem = DFunc_Tooltip:SetInventoryItem("player", slotid);
         if ( foundItem ) then
            tmp = getglobal("DFunc_TooltipTextLeft"..1);
            tmpLine = tmp:GetText();
            tmpLine = string.lower(tmpLine);
            if tmpLine then
               tmpGump[ktr] = tmpLine;
            end            
         end         
      end
      return tmpGump;
   end;

   EquipItem = function(invItem, gumpSlot)
      -- -------------------------------------------------------------------------------------
      -- This equipes the named item in the given slot.
      -- -------------------------------------------------------------------------------------
      for i = 0, 4, 1 do
         invItem = string.lower(invItem);
         maxSlots = GetContainerNumSlots(i)
         for ktr = 0, maxSlots, 1 do
            local texture, itemCount, locked, quality, readable = GetContainerItemInfo(1,2);
   	    linktext = GetContainerItemLink(i, ktr);
            if linktext then
               local _,_,name = string.find(linktext, "^.*%[(.*)%].*$");
               name=string.lower(name);
               if name == invItem then 
                  PickupContainerItem( i, ktr);
                  EquipCursorItem( GetInventorySlotInfo(gumpSlot.."Slot") )       
                  return
               end
            end
        end
      end
   end;

   
   LoadWardrobe = function( incTable )
      -- -------------------------------------------------------------------------------------
      -- This function equips the items in the passed table
      -- -------------------------------------------------------------------------------------
      
      -- First turn off the function that checks for +heal and mana/5 (to speed things up)
      DFUNCNoScan = true
   
      if UnitAffectingCombat("player") or DFUNCWeapon then
         -- Player is in combat so only equip weapons and ranged items
         Druid.Equip(incTable[16], incTable[17], incTable[18]);
      else
            -- Player is not in combat so equip everything
         for i, slotName in DFUNC_Slots do
            if (incTable[i] ~= Nil) and (incTable[i] ~= "") then
               Druid.EquipItem(incTable[i], slotName)
            end
         end
      end
      
      -- We've loaded everything so turn the +heal and mana/5 checking back on and recheck everything
      DFUNCNoScan = false
      Druid.ScanInventory()
   end;
      
   CheckWeapon = function ()
      -- This function merely returns the name of the weapon the player is wielding in his main hand
      local linktext = GetInventoryItemLink("player",GetInventorySlotInfo("MainHandSlot"))      
      local weaponInfo = " "
      if (linktext ~= nil) then
         local _,_,name = string.find(linktext, "^.*%[(.*)%].*$");
         weaponInfo = name;
      end      
      return weaponInfo
   end;

   Swap = function ( left1, right1, left2, right2 )
      -- Swaps weapons between two given sets
      if string.lower(Druid.CheckWeapon()) == string.lower(left1) then
         Druid.Equip(left2, right2);
      else
         Druid.Equip(left1, right1);
      end
   end;
   
   LowManaSwap = function ( Percent )
      -- This function checks the mana and combat conditions for low mana swaps. 
      if not(DFUNCSwap) or (DFUNCManaSet == Nil) then
         return
      end
      if (DFUNCManaSet[16] == Nil) then 
         return
      end
      if UnitAffectingCombat("player") then
         if not Druid.isUnitBuffUp("Player", "Nature_Lightning") then
            local pmana = floor(UnitMana( "Player" ) / UnitManaMax( "Player" ) * 100);
            if ( pmana <= Percent ) then               
               if string.lower(Druid.CheckWeapon()) ~= string.lower( DFUNCManaSet[16] ) then
                  Druid.LoadWardrobe( DFUNCManaSet )
               end
            end
         end   
      end      
   end;

   Equip = function ( LeftHand, RightHand, Idol )
      -- This function equips the sepcified weapons/ranged items.
      local found = false;
      local foundbag = 0;
      local foundslot = 0;
      local found2 = false;
      local foundbag2 = 0;
      local foundslot2 = 0;
      local found3 = false;
      local foundbag3 = 0;
      local foundslot3= 0;
      
      -- change strings to lower case
      if (LeftHand == Nil) then
         -- Nothing was passed so do nothing
         return
      else
         LeftHand = string.lower(LeftHand)
      end;
      if (RightHand ~= Nil) then
         RightHand = string.lower(RightHand)
      end
      if (Idol ~= Nil) then
         Idol = string.lower(Idol)
      end
      
      -- Check all the imtes in all the bags.
      for i = 0, 4, 1 do
         maxSlots = GetContainerNumSlots(i)
         for ktr = 0, maxSlots, 1 do
            local texture, itemCount, locked, quality, readable = GetContainerItemInfo(1,2);
   	    linktext = GetContainerItemLink(i, ktr);
            if linktext then
               local _,_,name = string.find(linktext, "^.*%[(.*)%].*$");
               name=string.lower(name);
               if name == LeftHand then 
                  found = true;
                  foundbag = i;
                  foundslot = ktr;
               end
               if RightHand ~= Nil then
                  if name == RightHand then
                     found2 = true;
                     foundbag2 = i;
                     foundslot2 = ktr;
                  end
               end
               if Idol ~= Nil then
                  if name == Idol then
                     found3 = true;
                     foundbag3 = i;
                     foundslot3 = ktr;
                  end
               end
            end
        end
      end
      if found then
         PickupContainerItem(foundbag, foundslot);
         AutoEquipCursorItem();         
      end;
      if found2 then
         PickupContainerItem(foundbag2, foundslot2);
         EquipCursorItem( 16 );         
      end;
      if found3 then
         PickupContainerItem(foundbag3, foundslot3);
         EquipCursorItem( 17 );
      end
   end;
   
   Outfit = function (msg)
      -- This is the / command handler for /doutfit 
      
      local findspace = string.find(string.lower(msg), " ");
      
      -- Parse the incomming message (The part after /doutfit)
      if findspace ~= nil then
         flag1 = string.sub(msg,1, findspace-1);
         temp = string.sub(msg, findspace+1).." ";
         findspace = string.find(temp," ");
         if findspace ~= nil then
            flag2 = string.sub(temp, 1, findspace-1);
            flag3 = string.sub(temp, findspace+1);
            flag3 = Druid.trim(flag3);
         else
            flag2 = temp;
            flag3 = "";
         end
      else
         flag1 = msg;
         flag2 = "";
         flag3 = "";
      end;
      flag1 = string.lower(flag1)
      flag2 = string.lower(flag2)
      flag3 = string.lower(flag3)
      if flag1 == "" then
         DEFAULT_CHAT_FRAME:AddMessage( "Usage: /doutfit [on/off/cat/bear/human/travel/aqua/mana/innervate/mount] [on/off]");
      end
      if flag1 == "off" then
         DFUNCSwap = false
         DEFAULT_CHAT_FRAME:AddMessage( "Automatic wardrobe changes have been disabled.");
         return
      end
      if flag1 == "weapon" then
         DFUNCWeapon = true
         DEFAULT_CHAT_FRAME:AddMessage( "Automatic wardrobe changes will only affect weapons/idols");
         return
      end
      if flag1 == "outfit" then
         DFUNCWeapon = false
         DEFAULT_CHAT_FRAME:AddMessage( "Automatic wardrobe changes will now affect all gear when not in combat");
         return
      end
      if flag1 == "on" then
         DFUNCSwap = true
         DEFAULT_CHAT_FRAME:AddMessage( "Automatic wardrobe changes have been enabled.");
      end         
      
      if flag1 == "cat" then
         if flag2 == "off" then
            DFUNCCatSet = {}
            DEFAULT_CHAT_FRAME:AddMessage( "Cat form wardobe is disabled");
         else
            DFUNCCatSet = Druid.ScanGump()
            DEFAULT_CHAT_FRAME:AddMessage( "Cat form wardrobe selection saved");
         end
      end

      if flag1 == "moon" then
         if flag2 == "off" then
            DFUNCMoonSet = {}
            DEFAULT_CHAT_FRAME:AddMessage( "Moonkin form wardobe is disabled");
         else
            DFUNCMoonSet = Druid.ScanGump()
            DEFAULT_CHAT_FRAME:AddMessage( "Moonkin form wardrobe selection saved");
         end
      end
      
      if flag1 == "travel" then
         if flag2 == "off" then
            DFUNCTravelSet = {}
            DEFAULT_CHAT_FRAME:AddMessage( "Travel form wardobe is disabled");
         else
            DFUNCTravelSet = Druid.ScanGump()
            DEFAULT_CHAT_FRAME:AddMessage( "Travel form wardrobe selection saved");
         end
      end

      if flag1 == "aqua" then
         if flag2 == "off" then
            DFUNCAquaSet = {}
            DEFAULT_CHAT_FRAME:AddMessage( "Aquatic form wardobe is disabled");
         else
            DFUNCAquaSet = Druid.ScanGump()
            DEFAULT_CHAT_FRAME:AddMessage( "Aquatic form wardrobe selection saved");
         end
      end
      
      if flag1 == "bear" then
         if flag2 == "off" then
            DFUNCBearSet = {}
            DEFAULT_CHAT_FRAME:AddMessage( "Bear form wardobe is disabled");
         else
            DFUNCBearSet = Druid.ScanGump()
            DEFAULT_CHAT_FRAME:AddMessage( "Bear form wardrobe selection saved");
         end
      end
      
      if flag1 == "human" then
         if flag2 == "off" then
            DFUNCHumanSet = {}
            DEFAULT_CHAT_FRAME:AddMessage( "Human form wardobe is disabled");
         else
            DFUNCHumanSet = Druid.ScanGump()
            DEFAULT_CHAT_FRAME:AddMessage( "Human form wardrobe selection saved");
         end
      end

      if flag1 == "mana" then
         if flag2 == "off" then
            DFUNCManaSet = {}
            DEFAULT_CHAT_FRAME:AddMessage( "Low Mana wardobe is disabled");
         else
            DFUNCManaSet = Druid.ScanGump()
            DEFAULT_CHAT_FRAME:AddMessage( "Low Mana wardrobe selection saved");
         end
      end
      
      if flag1 == "innervate" then
         if flag2 == "off" then
            DFUNCISet = {}
            DEFAULT_CHAT_FRAME:AddMessage( "Innervate wardobe is disabled");
         else
            DFUNCISet = Druid.ScanGump()
            DEFAULT_CHAT_FRAME:AddMessage( "Innervate wardrobe selection saved");
         end
      end
      
      if flag1 == "mount" then
         if flag2 == "off" then
            DFUNCMountSet = {}
            DEFAULT_CHAT_FRAME:AddMessage( "Mount wardobe is disabled");
         else
            DFUNCMountSet = Druid.ScanGump()
            DEFAULT_CHAT_FRAME:AddMessage( "Mount wardrobe selection saved");
         end
      end
      
   end;      

};

-- -------------------------------------------------------------------------------------
-- End DruidFunc
-- -------------------------------------------------------------------------------------
